package context;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.Policy.Parameters;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.batik.css.engine.value.StringValue;
import org.apache.commons.io.FilenameUtils;
import org.geotools.data.shapefile.ShapefileDataStore;
import org.geotools.data.simple.SimpleFeatureIterator;
import org.geotools.feature.FeatureIterator;
import org.jdom.Document;
import org.jdom.Element;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import micro.Agent;
import micro.InputParser;
import micro.ParametersManager;
import micro.Utils;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;

import environment.GillnetFishery;
import environment.Hexagon;
import environment.RapidFishery;
import environment.Region;
import environment.Sanctuary;
import environment.Stations;
import environment.TupleSpace;
import repast.simphony.context.Context;
import repast.simphony.context.DefaultContext;
import repast.simphony.context.space.gis.GeographyFactory;
import repast.simphony.context.space.gis.GeographyFactoryFinder;
import repast.simphony.dataLoader.ContextBuilder;
import repast.simphony.dataLoader.engine.ContextBuilderFactory;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.gis.Geography;
import repast.simphony.space.gis.GeographyParameters;
import repast.simphony.ui.RSApplication;
import repast.simphony.util.ContextUtils;
import repast.simphony.util.collections.IndexedIterable;
import repast.simphony.visualization.gis3D.WWUtils;

import java.io.FileWriter;
import java.io.IOException;

import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;

import org.jdom.Attribute;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.vividsolutions.jts.geom.Coordinate;

public class ContextMain extends DefaultContext implements ContextBuilder{

	private Context context;
	private Geography geography;

	//.shp is the shapefile extension 
	public final String fileExtension = ".shp";
	public final static File mainFolder = new File("shapefiles/grid");
	
	private final static String fileMigrationRules = "misc/migrationRules.csv";

	private final static String fisheryFilePath = "misc/mortalitiesProbabilities.csv";


	public String getFileExtension() {
		return fileExtension;
	}

	public File getMainFolder() {
		return mainFolder;
	}



	public Context build(Context context) {

		GeographyParameters geographyParameters = new GeographyParameters();
		Geography geography = GeographyFactoryFinder.createGeographyFactory(null)
				.createGeography("Geography", context, geographyParameters);
		GeometryFactory factory = new GeometryFactory();
		int firstYear = 0;
		int lastYear = 0;
		int numberOfYears = 0;
		List<Integer> yearsList = new ArrayList<Integer>();
		if (getMainFolder().isDirectory()){
			if (getMainFolder().list().length > 0){

				String[] folderList = mainFolder.list(new FilenameFilter() {

					@Override
					public boolean accept(File folder, String name) {

						return new File(folder, name).isDirectory();
					}
				});

				numberOfYears = folderList.length;
				for ( String year : folderList){
					yearsList.add((Integer.parseInt(year)));
				}

				firstYear = Integer.parseInt(folderList[0]);

				if(folderList.length == 1){
					lastYear = Integer.parseInt(folderList[0]);
				}else{
					lastYear = Integer.parseInt(folderList[folderList.length-1]);
				}

				//saving years data in an agent
				String id = "master";
				Kronos kronos = new Kronos(id, firstYear, lastYear, numberOfYears, yearsList);
				context.add(kronos);
				String path = getMainFolder()+"/"+folderList[0];
				File firstYearFolder = new File(path);
				if(firstYearFolder.isDirectory() && firstYearFolder.list().length > 0){
					//here call to load the hex grid 
					File[] files = firstYearFolder.listFiles(new FilenameFilter() {

						@Override
						public boolean accept(File firstYearFolder, String name) {

							return name.toLowerCase().endsWith(getFileExtension());
						}
					});

					if(files.length > 1 ){ System.out.println("INFO: The first year folder"
							+ "is containing more than one hexagonal grid shapefile ");}

					try {
						loadHexagonalGrid(files[0].toURI().toURL(), context, geography);
					} catch (MalformedURLException e) {

						System.out.println("ERROR: At loading the hexagonal grid shapefile due to shapefiles.");
						e.printStackTrace();
					}

				}else{

					System.out.println("INFO: The main shapefiles folder  is empty!");

				}

			}else{

				System.out.println("ERROR: The main shapefiles folder is not a folder.");

			}
		}


		//Adding the Region Agents for the soles presence. 
		URL zonePopulationZero = null;
		try {
			zonePopulationZero = new File("shapefiles/zones/solea 0/solea0.shp").toURL();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			System.out.println("INFO: Shapefile loaded for the population region zero is does not exist in the folder");
			e.printStackTrace();
		}
		final int populationType0 = 0;
		loadZones(zonePopulationZero, context, geography, populationType0);

		URL zonePopulationOne = null;
		try {
			zonePopulationOne = new File("shapefiles/zones/solea 1/solea1.shp").toURL();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			System.out.println("INFO: Shapefile loaded for the population region one is does not exist in the folder");
			e.printStackTrace();
		}
		final int populationType1 = 1;
		loadZones(zonePopulationOne, context, geography, populationType1);

		URL zonePopulationTwo = null;
		try {
			zonePopulationTwo = new File("shapefiles/zones/solea 2/solea2.shp").toURL();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			System.out.println("INFO: Shapefile loaded for the population region two is does not exist in the folder");
			e.printStackTrace();
		}
		final int populationType2 = 2;
		loadZones(zonePopulationTwo, context, geography, populationType2);

		URL zonePopulationThree = null;
		try {
			zonePopulationThree = new File("shapefiles/zones/solea 3/solea3.shp").toURL();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			System.out.println("INFO: Shapefile loaded for the population region three is does not exist in the folder");
			e.printStackTrace();
		}
		final int populationType3 = 3;
		loadZones(zonePopulationThree, context, geography, populationType3);

		URL zonePopulationFourth = null;
		try {
			zonePopulationFourth = new File("shapefiles/zones/solea 4/solea4.shp").toURL();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			System.out.println("INFO: Shapefile loaded for the population region fourth is does not exist in the folder");
			e.printStackTrace();
		}
		final int populationType4 = 4;
		loadZones(zonePopulationFourth, context, geography, populationType4);

		URL zonePopulationFifth = null;
		try {
			zonePopulationFifth = new File("shapefiles/zones/solea 5/solea5+.shp").toURL();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			System.out.println("INFO: Shapefile loaded for the population region fifth is does not exist in the folder");
			e.printStackTrace();
		}
		final int populationType5 = 5;
		loadZones(zonePopulationFifth, context, geography, populationType5);

		//Check and load sanctuary 
		File folder = new File("shapefiles/sanctuaries");//check if there is something on the folder
		final String fileExtension = ".shp";

		if (folder.isDirectory()){

			if (folder.list().length>0){

				File[] files = folder.listFiles(new FilenameFilter() {

					@Override
					public boolean accept(File folder, String name) {

						return name.toLowerCase().endsWith(fileExtension);
					}
				});
				for ( int x = 0 ; x < files.length ; x++){
					try {
						loadSanctuaries(files[x].toURI().toURL(), context, geography);
					} catch (MalformedURLException e) {

						System.out.println("ERROR: Error loading the sanctuaries zones due to the files.");
						e.printStackTrace();
					}
				}

			}else{

				System.out.println("INFO: Sanctuarie's directory is empty!");

			}

		}else{

			System.out.println("ERROR: The directory for the sanctuary folder, is not a folder or it does not exist.");

		}

		//Loading the rapid fishery's shape-files 
		File folderFisheriesRapid = new File("shapefiles/fisheries/rapid");//check if there is something on the folder

		if (folderFisheriesRapid.isDirectory()){

			if (folderFisheriesRapid.list().length>0){

				File[] files = folderFisheriesRapid.listFiles(new FilenameFilter() {

					@Override
					public boolean accept(File folder, String name) {

						return name.toLowerCase().endsWith(fileExtension);
					}
				});
				for ( int x = 0 ; x < files.length ; x++){
					try {
						loadFisheries(files[x].toURI().toURL(), context, geography);
					} catch (MalformedURLException e) {

						System.out.println("ERROR: Error loading the rapid fisheries zones due to the files.");
						e.printStackTrace();
					}
				}

			}else{

				System.out.println("INFO: Rapid Fisheries directory is empty!");

			}

		}else{

			System.out.println("ERROR: The directory for the rapid fishery folder, is not a folder.");

		}

		//Loading the Gillnet fishery's shape-files 
		File folderFisheriesGillnet = new File("shapefiles/fisheries/gillnet");//check if there is something on the folder

		if (folderFisheriesGillnet.isDirectory()){

			if (folderFisheriesGillnet.list().length>0){

				File[] files = folderFisheriesGillnet.listFiles(new FilenameFilter() {

					@Override
					public boolean accept(File folder, String name) {

						return name.toLowerCase().endsWith(fileExtension);
					}
				});
				for ( int x = 0 ; x < files.length ; x++){
					try {
						loadFisheriesGillnet(files[x].toURI().toURL(), context, geography);
					} catch (MalformedURLException e) {

						System.out.println("ERROR: Error loading the Gillnet fisheries zones due to the files.");
						e.printStackTrace();
					}
				}



			}else{

				System.out.println("INFO: Gillnet Fisheries directory is empty!");

			}

		}else{

			System.out.println("ERROR: The Gillnet directory for the fishery folder, is not a folder.");

		}


		Ares ares = new Ares(this.fisheryFilePath);
		context.add(ares);
		Emigrator emigrator = new Emigrator(this.fileMigrationRules);
		context.add(emigrator);
		return context;
	}

	public void loadZones(URL url, Context context, Geography geography, int populationType){

		SimpleFeatureIterator fiter = null;
		ShapefileDataStore store = null;
		try {
			store = new ShapefileDataStore(url);
		} catch (MalformedURLException e) {
			e.printStackTrace();
		}

		CoordinateReferenceSystem crs = null;
		try {
			crs = store.getSchema().getCoordinateReferenceSystem();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		try {
			fiter = store.getFeatureSource().getFeatures().features();
		} catch (IOException e) {
			e.printStackTrace();
		}

		while(fiter.hasNext()){
			SimpleFeature feature = fiter.next();
			Geometry geom = (Geometry)feature.getDefaultGeometry();

			if (geom instanceof MultiPolygon){
				MultiPolygon mp = (MultiPolygon)feature.getDefaultGeometry();
				geom = (Polygon)mp.getGeometryN(0);


				int value = (Integer)feature.getAttribute("VALUE");

				Region region = new Region();

				region.setValue(value);
				region.setPopulationType(populationType);
				
				context.add(region);

				if (crs != null){
					WWUtils.projectGeometryToWGS84(geom, crs);
				}

				geography.move(region, geom);
			}
			else{
				System.out.println("INFO: NULL agent for population reagion:  " + geom);
			}

		}

	}

	public void loadSanctuaries(URL url, Context context, Geography geography){

		SimpleFeatureIterator fiter = null;
		ShapefileDataStore store = null;
		try {
			store = new ShapefileDataStore(url);
		} catch (MalformedURLException e) {
			e.printStackTrace();
		}

		CoordinateReferenceSystem crs = null;
		try {
			crs = store.getSchema().getCoordinateReferenceSystem();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		try {
			fiter = store.getFeatureSource().getFeatures().features();
		} catch (IOException e) {
			e.printStackTrace();
		}

		while(fiter.hasNext()){
			SimpleFeature feature = fiter.next();
			Geometry geom = (Geometry)feature.getDefaultGeometry();

			if (geom instanceof MultiPolygon){
				MultiPolygon mp = (MultiPolygon)feature.getDefaultGeometry();
				geom = (Polygon)mp.getGeometryN(0);
				Double latitude = 0.0;
				Double longitude = 0.0;
				boolean coordPresence = false;
				double ID = (Double)feature.getAttribute("ide");
				int id = (int) Math.round(ID) ;
				if(feature.getAttribute("latitude") != null && feature.getAttribute("longitude") != null){
					latitude = (Double) feature.getAttribute("latitude");
					longitude = (Double) feature.getAttribute("longitude");
					coordPresence = true;
				}
				Sanctuary hexSanctuary = new Sanctuary();
				hexSanctuary.setID(id);
				if(coordPresence){
					hexSanctuary.setLatitude(latitude);
					hexSanctuary.setLongitude(longitude);
				}

				context.add(hexSanctuary);

				if (crs != null){
					WWUtils.projectGeometryToWGS84(geom, crs);
				}

				geography.move(hexSanctuary, geom);


			}
			else{
				System.out.println("INFO: NULL agent for sanctuary hexagon:  " + geom);
			}

		}

	}

	//load Rapid fishery shape files 
	public void loadFisheries(URL url, Context context, Geography geography){

		SimpleFeatureIterator fiter = null;
		ShapefileDataStore store = null;
		try {
			store = new ShapefileDataStore(url);
		} catch (MalformedURLException e) {
			e.printStackTrace();
		}

		CoordinateReferenceSystem crs = null;
		try {
			crs = store.getSchema().getCoordinateReferenceSystem();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		try {
			fiter = store.getFeatureSource().getFeatures().features();
		} catch (IOException e) {
			e.printStackTrace();
		}

		while(fiter.hasNext()){
			SimpleFeature feature = fiter.next();
			Geometry geom = (Geometry)feature.getDefaultGeometry();

			if (geom instanceof MultiPolygon){
				MultiPolygon mp = (MultiPolygon)feature.getDefaultGeometry();
				geom = (Polygon)mp.getGeometryN(0);

				Long ID = (Long)feature.getAttribute("ID");
				Long total = (Long)feature.getAttribute("TOTAL");

				if(total != 0){
					RapidFishery rapidPoligon = new RapidFishery();

					rapidPoligon.setId(ID);
					rapidPoligon.setTotal(total);

					context.add(rapidPoligon);
					if (crs != null){
						WWUtils.projectGeometryToWGS84(geom, crs);
					}

					geography.move(rapidPoligon, geom);

				}
			}
			else{
				System.out.println("INFO: NULL agent for rapid fishery hexagon:  " + geom);
			}

		}

	}

	//loading gillnets fishery shapefiles 
	public void loadFisheriesGillnet(URL url, Context context, Geography geography){

		SimpleFeatureIterator fiter = null;
		ShapefileDataStore store = null;
		try {
			store = new ShapefileDataStore(url);
		} catch (MalformedURLException e) {
			e.printStackTrace();
		}

		CoordinateReferenceSystem crs = null;
		try {
			crs = store.getSchema().getCoordinateReferenceSystem();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		try {
			fiter = store.getFeatureSource().getFeatures().features();
		} catch (IOException e) {
			e.printStackTrace();
		}

		while(fiter.hasNext()){
			SimpleFeature feature = fiter.next();
			Geometry geom = (Geometry)feature.getDefaultGeometry();

			if (geom instanceof MultiPolygon){
				MultiPolygon mp = (MultiPolygon)feature.getDefaultGeometry();
				geom = (Polygon)mp.getGeometryN(0);

				Long ID = (Long)feature.getAttribute("ID");
				Long gridCode = (Long)feature.getAttribute("GRIDCODE");

				GillnetFishery gillnettingPoligon = new GillnetFishery();

				gillnettingPoligon.setId(ID);
				gillnettingPoligon.setGridCode(gridCode);

				context.add(gillnettingPoligon);
				if (crs != null){
					WWUtils.projectGeometryToWGS84(geom, crs);
				}

				geography.move(gillnettingPoligon, geom);


			}
			else{
				System.out.println("INFO: NULL agent for rapid fishery hexagon:  " + geom);
			}

		}

	}


	public void loadHexagonalGrid(URL url, Context context, Geography geography){

		SimpleFeatureIterator fiter = null;
		ShapefileDataStore store = null;

		try {
			store = new ShapefileDataStore(url);
		} catch (MalformedURLException e) {
			e.printStackTrace();
		}

		CoordinateReferenceSystem crs = null;
		try {
			crs = store.getSchema().getCoordinateReferenceSystem();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		try {
			fiter = store.getFeatureSource().getFeatures().features();
		} catch (IOException e) {
			e.printStackTrace();
		}

		while(fiter.hasNext()){
			SimpleFeature feature = fiter.next();
			Geometry geom = (Geometry)feature.getDefaultGeometry();

			if (geom instanceof MultiPolygon){

				MultiPolygon mp = (MultiPolygon)feature.getDefaultGeometry();
				geom = (Polygon)mp.getGeometryN(0);

				Double ID = (Double) feature.getAttribute("ide");
				Double centroidLatitude = (Double)feature.getAttribute("latitude");
				Double centroidLongitude = (Double)feature.getAttribute("longitude");
				Double k = (Double)feature.getAttribute("K");

				Double soleZero = 0.0;
				if (feature.getAttribute("kr_0_max") != null){
					soleZero = (Double) feature.getAttribute("kr_0_max");;
				}else if(feature.getAttribute("kr_0_max") == null){
					soleZero = 0.0;
				}

				Double soleOne = 0.0;
				if (feature.getAttribute("kr_1_max") != null){
					soleOne = (Double) feature.getAttribute("kr_1_max");
				}else if(feature.getAttribute("kr_1_max") == null){
					soleOne = 0.0;
				}

				Double soleTwo = 0.0; 
				if (feature.getAttribute("kr_2_max") != null){
					soleTwo = (Double) feature.getAttribute("kr_2_max");
				}else if(feature.getAttribute("kr_2_max") == null){
					soleTwo = 0.0;
				}

				Double soleThree = 0.0;
				if (feature.getAttribute("kr_3_max") != null){
					soleThree = (Double) feature.getAttribute("kr_3_max");
				}else if(feature.getAttribute("kr_3_max") == null){
					soleThree = 0.0;
				}

				Double soleFour = 0.0; 
				if(feature.getAttribute("kr_4_max") != null){
					soleFour = 	(Double) feature.getAttribute("kr_4_max");
				}else if(feature.getAttribute("kr_4_max") == null){
					soleFour = 0.0;
				}

				Double soleFive = 0.0; 
				if(feature.getAttribute("kr_5_max") != null){
					soleFive = (Double) feature.getAttribute("kr_5_max");
				}else if(feature.getAttribute("kr_5_max") == null){
					soleFive = 0.0;
				}

				Double seaFlag = (Double) feature.getAttribute("Sea");
				String sea = "IT";		
				if (seaFlag == 1.0) {
					sea = "CRO";
				}
				
				//hexagon directions North, south, east, west
				Double north = (Double) feature.getAttribute("North");
				boolean North = false;
				if (north == 1.0 ) {
					North = true;
				}
				Double south = (Double) feature.getAttribute("South");
				boolean South = false;
				if (south == 1.0) {
					South = true;
				}
				Double east = (Double) feature.getAttribute("East");
				boolean East = false;
				if (east == 1.0) {
					East = true;
				}
				Double west = (Double) feature.getAttribute("West");
				boolean West = false;
				if (west == 1.0) {
					West = true;
				}
				
				if(soleZero < 0){
					soleZero = 0.0;
				}
				if(soleOne < 0){
					soleOne = 0.0;
				}
				if(soleTwo < 0){
					soleTwo = 0.0;
				}
				if(soleThree < 0){
					soleThree = 0.0;
				}
				if(soleFour < 0){
					soleFour = 0.0;
				}
				if(soleFive < 0){
					soleFive = 0.0;
				}
				int id = (int) Math.round(ID);

				Hexagon hexagon = new Hexagon();
				hexagon.setID(id);
				hexagon.setCentroidLatitude(centroidLatitude);
				hexagon.setCentroidLongitude(centroidLongitude);
				hexagon.setK(k);

				hexagon.setPopulation(soleZero, 0);
				hexagon.setPopulation(soleOne, 1);
				hexagon.setPopulation(soleTwo, 2);
				hexagon.setPopulation(soleThree, 3);
				hexagon.setPopulation(soleFour, 4);
				hexagon.setPopulation(soleFive, 5);
				hexagon.setSea(sea);
				hexagon.setNorth(North);
				hexagon.setSouth(South);
				hexagon.setEast(East);
				hexagon.setWest(West);
				//create cohort
				double[] cohort = new double[6];
				cohort[0] = soleZero;
				cohort[1] = soleOne;
				cohort[2] = soleTwo;
				cohort[3] = soleThree;
				cohort[4] = soleFour;
				cohort[5] = soleFive;
				
				File file = new File(url.getPath());
				String val = file.getParentFile().getName();
				int year = Integer.valueOf(val);
				hexagon.setPopulationCohort(year, cohort);
				hexagon.setCohortTimeInitial(year);
				
				context.add(hexagon);

				if (crs != null){
					WWUtils.projectGeometryToWGS84(geom, crs);
				}

				geography.move(hexagon, geom);

			}
			else{
				System.out.println("INFO: NULL agent hexagon (geometry):  " + geom);
			}

		}
		//add the grid year as visited 
		IndexedIterable<Kronos> iter = context.getObjects(Kronos.class);
		if (iter.size() == 1) {
			Object obj = iter.get(0);
			Kronos timer = (Kronos) obj;
			File file = new File(url.getFile());
			String folderName = file.getAbsoluteFile().getParentFile().getName() ;
			timer.setYearVisited(Integer.parseInt(folderName));

		}else {
			System.out.println("INFO: Exist more than one chronos agent");
		}

	}
	
}
