package context;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.lang.instrument.Instrumentation;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.swing.JOptionPane;

import org.geotools.data.shapefile.ShapefileDataStore;
import org.geotools.data.simple.SimpleFeatureIterator;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;

import environment.Hexagon;
import environment.Region;
import environment.Stations;
import micro.ParametersManager;
import repast.simphony.context.Context;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunEnvironmentBuilder;
import repast.simphony.engine.environment.Runner;
import repast.simphony.engine.schedule.IScheduleFactory;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.parameter.Parameters;
import repast.simphony.space.gis.Geography;
import repast.simphony.util.ContextUtils;
import repast.simphony.util.collections.IndexedIterable;
import repast.simphony.visualization.gis3D.WWUtils;

public class Kronos {

	public List<Integer> getYearsList() {
		return yearsList;
	}

	public void setYearsList(List<Integer> yearsList) {
		this.yearsList = yearsList;
	}

	public int getFirstYear() {
		return firstYear;
	}

	public void setFirstYear(int firstYear) {
		this.firstYear = firstYear;
	}

	public int getLastYear() {
		return lastYear;
	}

	public void setLastYear(int lastYear) {
		this.lastYear = lastYear;
	}

	public int getNumberOfYears() {
		return numberOfYears;
	}

	public void setNumberOfYears(int numberOfYears) {
		this.numberOfYears = numberOfYears;
	}

	
	public Map<Integer, Boolean> getYearVisited() {
		return yearVisited;
	}

	public boolean isTheYearVisited(int year){
		return this.getYearVisited().get(year);
	}

	public void setYearVisited(Map<Integer, Boolean> yearVisited) {
		this.yearVisited = yearVisited;
	}

	public void setYearVisited(int year){
		this.yearVisited.put(year, true);
	}

	private String ID; 
	private int firstYear;
	private int lastYear;
	private int numberOfYears;
	private List<Integer> yearsList;
	private Map<Integer,Boolean> yearVisited = new HashMap<Integer,Boolean>();

	public Kronos(String iD, int firstYear, int lastYear, int numberOfYears,
			List<Integer> yearsList) {
		super();
		ID = iD;
		this.firstYear = firstYear;
		this.lastYear = lastYear;
		this.numberOfYears = numberOfYears;
		this.yearsList = yearsList;
		this.yearVisited = new HashMap<Integer,Boolean>();
		initializateYearVisited();
	}


	private void initializateYearVisited() {
		for (int i = 0; i < yearsList.size(); i++) {
			int year = yearsList.get(i);
			this.yearVisited.put(year, false);
		}

	}

	public int getMonthFromTick(){
		double month = 0; 

		final int tickSimulationStartsAt = 2;
		int firstYear = 0;
		int tick  = (int)RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
		month = ((tick - tickSimulationStartsAt)/2%12) %24 ;
		if (month > 0) {
			return (int) month;
		}else {
			//System.out.println("ERROR: Month cannot be negative");
			return 0;
		}
		//I could convert it to some kind of string "JAN" for 0 and son on... if it is needed 
		//return (int) month;
	}

	public int getYearFomTick(){
		int year = 0;
		final int tickSimulationStartsAt = 2;
		int firstYear = 0;
		int tick  = (int)RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
		Context context = ContextUtils.getContext(this);
		IndexedIterable<Kronos> iterator = context.getObjects(Kronos.class);
		if(iterator.size() == 1){
			for (Object timer : iterator) {
				Kronos kronos = (Kronos) timer;
				firstYear = kronos.firstYear;
			}
		}
		year = firstYear + (tick - tickSimulationStartsAt) /24;

		return year;
	}

	/*
	 * The method updatingHexagonalGrid adds the newborns data for each year
	 * to hexagon. 
	 * */
	@ScheduledMethod(start = 26 , interval = 24 , priority = 0)
	public void updatingHexagonalGrid(){
		Context context = ContextUtils.getContext(this);
		int currentYear = getYearFomTick();
		IndexedIterable<Hexagon> iterator = context.getObjects(Hexagon.class);
		IndexedIterable<Stations> iterStations = context.getObjects(Stations.class);
		Map<Integer, Double> newBornsClass0 = new HashMap<Integer, Double>();
		Map<Integer, double[]> coordinateID = new HashMap<Integer, double[]>();
		List<Hexagon> hexUpdated = new ArrayList<Hexagon>();
		//check time and that year exist in the available data

		//hexagons to be updated
		for (Hexagon hex : iterator) {
			if (hex.getClassesPresent()[0] == true) {
				hexUpdated.add(hex);
			}
		}

		if (yearsList.contains(currentYear)) {
			if (!yearVisited.get(currentYear)) {
				File folderPath = new File(ContextMain.mainFolder+"/"+currentYear);

				//add the grid --> get the information for newborns
				File[] files = null;
				if(folderPath.isDirectory() && folderPath.list().length > 0){
					//here call to load the hex grid 
					files = folderPath.listFiles(new FilenameFilter() {

						@Override
						public boolean accept(File firstYearFolder, String name) {

							return name.toLowerCase().endsWith(".shp");
						}
					});

					if(files.length > 1 ){ System.out.println("INFO: The first year folder"
							+ "is containing more than one hexagonal grid shapefile ");}


				}else{

					System.out.println("INFO: The shapefiles folder  is empty for year: " +currentYear);

				}

				URL url = null;
				try {
					url = files[0].toURI().toURL();
				} catch (MalformedURLException exp) {
					// TODO Auto-generated catch block
					exp.printStackTrace();
				}

				SimpleFeatureIterator fiter = null;
				ShapefileDataStore store = null;

				try {
					store = new ShapefileDataStore(url);
				} catch (MalformedURLException e) {
					e.printStackTrace();
				}

				CoordinateReferenceSystem crs = null;
				try {
					crs = store.getSchema().getCoordinateReferenceSystem();
				} catch (IOException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}

				try {
					fiter = store.getFeatureSource().getFeatures().features();
				} catch (IOException e) {
					e.printStackTrace();
				}

				while(fiter.hasNext()){
					SimpleFeature feature = fiter.next();
					Geometry geom = (Geometry)feature.getDefaultGeometry();
					//double [] popToAdd = new double[6];
					if (geom instanceof MultiPolygon){

						MultiPolygon mp = (MultiPolygon)feature.getDefaultGeometry();
						geom = (Polygon)mp.getGeometryN(0);
						//be careful here because the id column has to be the same as the in the main context
						Double ID = (Double) feature.getAttribute("ide");						
						Double soleZero = 0.0;
						Double centroidLatitude = (Double)feature.getAttribute("latitude");
						Double centroidLongitude = (Double)feature.getAttribute("longitude");
						if (feature.getAttribute("kr_0_mean") != null){
							soleZero = (Double) feature.getAttribute("kr_0_mean");
						}else if(feature.getAttribute("kr_0_mean") == null){
							soleZero = 0.0;
						}

						if(soleZero < 0 ){
							soleZero = 0.0;
						}

						int id = (int) Math.round(ID);
						double[] coordinates = {centroidLatitude , centroidLongitude}; 
						newBornsClass0.put(id,soleZero);
						double code = centroidLatitude + centroidLongitude;
						//						newBornsClassZero.put(code, soleZero);
						coordinateID.put(id, coordinates);



					}
					else{
						System.out.println("INFO: NULL agent the new hexagon (geometry) that you try to add:  " + geom);
					}

				}

				//updating the hexagon class zero population, just those that are in the list
				try {
					for (Hexagon hexagon : hexUpdated) {//hexUpdated is list with the hexagons that has the sole age zero 
						if (newBornsClass0.get(hexagon.getID()) != null && 
								!hexagon.isStation()) {
							double newborns = newBornsClass0.get(hexagon.getID());
							if (newborns > 0.0) {
								double[] cohortPopulation = { newborns, 0.0, 0.0, 0.0, 0.0, 0.0 };
								hexagon.setPopulationCohort(getYearFomTick(), cohortPopulation);
								hexagon.setNewCohortTime(getYearFomTick());
							}
							//check if it is a station
						
						}else if (newBornsClass0.get(hexagon.getID()) != null && 
								hexagon.isStation()) {
							for (Stations station : iterStations) {
								if (station.getName() == hexagon.getNameStation()) {
									double newborns = station.getSole0();
									if (newborns > 0.0) {
										double[] cohortPopulation = { newborns, 0.0, 0.0, 0.0, 0.0, 0.0 };
										hexagon.setPopulationCohort(getYearFomTick(), cohortPopulation);
										hexagon.setNewCohortTime(getYearFomTick());
									}
								}
							}	
						}
					}
				} catch (Exception exp) {
					// TODO Auto-generated catch block
					System.out.println("ERROR: Like in MS (unmentionable)... something went wrong. Check the newborns update. Ex: " + exp);
					exp.printStackTrace();
				}
				//adding the folder year to the visited list 
				setYearVisited(currentYear);
				//System.out.println("added year: " + currentYear);
			}else {
				System.out.println("ERROR: year that are you trying to add, was already added.");
			}
		}else {
			//here will comes the simulation code for the XSA
			System.out.println("INFO: Data year does not exist in the yearsList to update the newborns for: "+ currentYear);
			RunEnvironment.getInstance().pauseRun();
			JOptionPane.showMessageDialog(null, "The simulation was paused due to data for newborns is finished", "Simulation Paused", JOptionPane.INFORMATION_MESSAGE);
		}

	}

	@ScheduledMethod(start = 1, interval = 24 , priority = 0)
	public void addStations(){
		Context context = ContextUtils.getContext(this);
		Geography geography = (Geography) context.getProjection("Geography");
		String path = "shapefiles/stations/"+getYearFomTick();
		File filePath = new File(path);
		if (filePath.exists()) {
			//CHECK IF THIS DOESNT EXIST DO NOTHING
			final String fileExtension = ".shp";
			URL stations = null;
			if (filePath.isDirectory()) {

				if (filePath.list().length > 0) {

					File[] files = filePath.listFiles(new FilenameFilter() {

						@Override
						public boolean accept(File folder, String name) {

							return name.toLowerCase().endsWith(fileExtension);
						}
					});
					//for ( int x = 0 ; x < files.length ; x++){
					try {
						//loadSanctuaries(files[x].toURI().toURL(), context, geography);
						stations = files[0].toURI().toURL();
					} catch (MalformedURLException e) {

						System.out
						.println("ERROR: Error loading stationes due to the files.");
						e.printStackTrace();
					}
					//}

				} else {

					System.out
					.println("INFO: the shapefiles folder of the stations is empty! ");

				}

			} else {

				System.out.println("ERROR: The directory for the points folder, is not a folder or it does not exist.");

			}
			
			SimpleFeatureIterator fiter = null;
			ShapefileDataStore store = null;
			try {
				store = new ShapefileDataStore(stations);
			} catch (MalformedURLException e) {
				e.printStackTrace();
			}
			CoordinateReferenceSystem crs = null;
			try {
				crs = store.getSchema().getCoordinateReferenceSystem();
			} catch (IOException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
			try {
				fiter = store.getFeatureSource().getFeatures().features();
			} catch (IOException e) {
				e.printStackTrace();
			}
			while (fiter.hasNext()) {
				SimpleFeature feature = fiter.next();
				Geometry geom = (Geometry) feature.getDefaultGeometry();

				if (geom instanceof Point) {

					geom = (Point) feature.getDefaultGeometry();

					long name = (Long) feature.getAttribute("F01_HAUL_N");
					double latitude = (Double) feature.getAttribute("LAT_S");
					double longitude = (Double) feature.getAttribute("LON_S");
					double popZero = (double) feature.getAttribute("SOLEVUL _0");
					double popOne = (Double) feature.getAttribute("SOLEVUL _1");
					double popTwo = (Double) feature.getAttribute("SOLEVUL _2");
					double popThree = (Double) feature.getAttribute("SOLEVUL _3");
					double popFour = (Double) feature.getAttribute("SOLEVUL _4");
					double popFive = (Double) feature.getAttribute("SOLEVUL _5");
					double total = (Double) feature.getAttribute("TOT");
					Stations station = new Stations();

					station.setName((int)name);
					station.setLat(latitude);
					station.setLon(longitude);
					station.setSole0(popZero);
					station.setSole1(popOne);
					station.setSole2(popTwo);
					station.setSole3(popThree);
					station.setSole4(popFour);
					station.setSole5Plus(popFive);
					station.setTotal(total);

					context.add(station);

					if (crs != null) {
						WWUtils.projectGeometryToWGS84(geom, crs);
					}

					geography.move(station, geom);

				} else {
					System.out.println("INFO: Null agent for the station:  "
							+ geom);
				}

			}
		}

	}

	//for testing purpose -------->>> this is what is executing now 
	@ScheduledMethod(start = 2 , priority = ScheduleParameters.FIRST_PRIORITY)
	public void printingHeader(){
		System.out.println("year;StationID;Sole0;Sole1;Sole2;Sole3;Sole4;Sole5;"+
				"AbundanceStation;HexSole0;HexSole1;HexSole2;HexSole3;HexSole4;HexSole5;AbundanceHex");
	}

	@ScheduledMethod(start = 2 , interval = 24, priority = ScheduleParameters.LAST_PRIORITY)
	public void dataCollector(){
		Context context = ContextUtils.getContext(this);
		IndexedIterable<Hexagon> iterator = context.getObjects(Hexagon.class);
		IndexedIterable<Stations> iteSta = context.getObjects(Stations.class);

		for (Hexagon hex : iterator) {
			for (Stations station : iteSta) {

				if (hex.isStation() == true && station.getName() == hex.getNameStation()) {
					//station data
					int name = station.getName();
					double staPop0 = station.getSole0();
					double staPop1 = station.getSole1();
					double staPop2 = station.getSole2();
					double staPop3 = station.getSole3();
					double staPop4 = station.getSole4();
					double staPop5 = station.getSole5Plus();
					double statTotal = station.getTotal();
					//hexagon data 
					double class0Population = hex.getClass0Population();
					double class1Population = hex.getClass1Population();
					double class2Population = hex.getClass2Population();
					double class3Population = hex.getClass3Population();
					double class4Population = hex.getClass4Population();
					double class5Population = hex.getClass5Population();
					double abundance = hex.getNumberOfIndividuals();

					System.out.println( getYearFomTick() +";" + name +";"+ staPop0 + ";"  + staPop1 + ";" + staPop2
							+";" + staPop3 +";" + staPop4 +";" + staPop5 +";" + statTotal +";" + 
							class0Population+";" + class1Population +";" + class2Population +";" + class3Population +";"
							+ class4Population+";"+ class5Population +";" + abundance+";");
				}

			}//else do nothing
		}
	}
	
	public double getTick(){
		return RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
	}
	
	@ScheduledMethod(start = 2, interval = 0 , priority = 1)
	public void updateHexagonStations(){
		Context context = ContextUtils.getContext(this);
		IndexedIterable<Hexagon> iterHexagon = context.getObjects(Hexagon.class);
		IndexedIterable<Stations> iterStation = context.getObjects(Stations.class);
		
		if (iterStation.size() > 0) {
			for (Hexagon hexagon : iterHexagon) {

				if (hexagon.isStation()) {
					//System.out.println("es una estacion: " + hexagon.getID());
					for (Stations station : iterStation) {
						if (hexagon.getNameStation() == station.getName()) {
							double soleZero = station.getSole0();
							double sole1 = station.getSole1();
							double sole2 = station.getSole2();
							double sole3 = station.getSole3();
							double sole4 = station.getSole4();
							double sole5 = station.getSole5Plus();

							double[] hexNewPop = hexagon.getCohort().get(2005);
							double newPopulation0 = soleZero /*+ hexNewPop[0]*/;
							double newPop1 = sole1 /*+ hexNewPop[1]*/;
							double newPop2 = sole2 /*+ hexNewPop[2]*/;
							double newPop3 = sole3 /*hexNewPop[3]*/;
							double newPop4 = sole4 /*+ hexNewPop[4]*/;
							double newPop5 = sole5 /*+ hexNewPop[5]*/;
							hexNewPop[0] = newPopulation0;
							hexNewPop[1] = newPop1;
							hexNewPop[2] = newPop2;
							hexNewPop[3] = newPop3;
							hexNewPop[4] = newPop4;
							hexNewPop[5] = newPop5;
							hexagon.getCohort().put(getFirstYear(), hexNewPop);
						}

					}
				}
			}
		}
	}

}



