package environment;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.ThreadLocalRandom;

import javolution.context.ConcurrentException;

import org.apache.commons.collections.functors.InstanceofPredicate;
import org.apache.commons.collections.iterators.FilterIterator;
import org.apache.commons.math3.distribution.BinomialDistribution;
import org.apache.commons.math3.distribution.NormalDistribution;
import org.apache.commons.math3.ode.sampling.StepNormalizer;
import org.apache.poi.ss.formula.functions.T;
import org.geotools.geometry.jts.JTS;
import org.geotools.referencing.GeodeticCalculator;
import org.jfree.data.function.NormalDistributionFunction2D;
import org.jscience.geography.coordinates.crs.CoordinateReferenceSystem;
import org.opengis.geometry.DirectPosition;
import org.opengis.referencing.operation.TransformException;
import org.stringtemplate.v4.compiler.STParser.element_return;
import org.stringtemplate.v4.compiler.STParser.mapExpr_return;

import micro.Dummy;
import micro.InputParser;
import micro.ParametersManager;
import repast.simphony.context.Context;
import repast.simphony.context.space.gis.GeographyFactoryFinder;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.parameter.Parameter;
import repast.simphony.parameter.RunParameterSetter;
import repast.simphony.query.space.gis.GeographyWithin;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.gis.Geography;
import repast.simphony.space.gis.GeographyParameters;
import repast.simphony.statecharts.runtime.MyAgent;
import repast.simphony.util.ContextUtils;
import repast.simphony.util.collections.IndexedIterable;
import cern.jet.random.Binomial;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.operation.distance.DistanceOp;
import com.vividsolutions.jts.operation.distance.GeometryLocation;

import context.Ares;
import context.Emigrator;
import context.Kronos;
import context.ContextMain;

public class Hexagon {

	//variables declaration 
	private int ID;
	private double centroidLatitude = 0.0;
	private double centroidLongitude = 0.0;
	private double k = 0.0;
	private double M = 0.0;
	private double F = 0.0;
	private boolean sanctuary = false;

	// both are the fishery mortality probability. Loaded for the presence of either rapid or gill fleets 
	private double fRapid = 0.0;
	private double fGill = 0.0;

	//after testing we can delet this 
	private boolean isStation;
	private double latStation;
	private double longStation;
	private int nameStation;

	//Distance to the neighbor in meters, it is used at BC  1500
	//I made the experiment extending the distance, with 4000 gets 16 neighbors. It seems that from 500 to 2000 is good
	//Although I have made some special cases in the hexagons... really particular cases... 
	private double distance = 1500;

	//constants for the neighbors directions
	private static final int precision = 11;//truncate to 11 positions in order to fix the the errors
	private static final double nordLatitude = round(-0.04711180393179859,precision);
	private static final double nordLongitude = 0.0;
	private static final double nordEstLatitude = round(-0.023555901965899295, precision);
	private static final double nordEstLongitude = round(-0.040800019023100376, precision);
	private static final double nordWestLatitude = round(-0.023555901965899295,precision);
	private static final double nordWestLongitude = round(0.04080001902299912,precision); 
	private static final double southLatitude = round(0.04711180393179859,precision);
	private static final double southLongitude = 0.0;
	private static final double southEstLatitude = round(0.023555901965899295,precision);
	private static final double southEstLongitude = round(-0.040800019023100376,precision);
	private static final double southWestLatitude = round(0.023555901965899295,precision);
	private static final double southWestLongitude = round(0.04080001902299912,precision);



	//neighbors directions 
	Map<Integer, String> neighborDirections = new HashMap<>();
	Map<String, Integer> neighboursDirections = new HashMap<>();

	//Different list and maps for the neighborhood 
	private List<Hexagon> neighbors = new ArrayList<Hexagon>();
	private List<Integer> idList = new ArrayList<>();
	//coordinate[0] = latitude -- coordinate[1] = longitude
	private Map<Integer, double[]> centroids = new HashMap<>();  

	//values of the classes
	private double[] population = new double[6];
	private Map<Hexagon, float[]> inIndividualsData;
	private Map<Hexagon, float[]> outIndividualsData;

	//array to say what population classes are in the hexagon 
	boolean[] classesPresent = new boolean [6];
	private boolean rapidPresence = false;
	private boolean gillPresence = false; 


	private double[] biomass = new double [6];

	private float[] populationEmigrated = new float[6];

	
	//Variable to look what fishery apply
	private String sea;
	//catches
	//stores per year per kind of fishery per hexagon kind
	Map<Integer, double[]> catchesRapidItaly = new HashMap<>();
	Map<Integer, double[]> catchesGillItaly = new HashMap<>();
	Map<Integer, double[]> catchesGillCroatia = new HashMap<>();

	double rapidItalyCatches = 0.0;
	double gillItalyCatches = 0.0;
	double gillCroatiaCatches = 0.0;

	//for the cohort study 
	private Map<Integer, double[]> cohort = new HashMap<>();
	private Map<Integer, Integer> cohortAge = new HashMap<>();
	//Nov-Dic-Jan-Feb-Mar-Apr-May-Jun-Jul-Ago-Sep-Oct
	private final int[] growthVector = {0, 0, 0, 100, 0, 0, 0, 0, 0, 0, 0, 0};

	//directions of the hexagon
	boolean north = false;
	boolean south = false;
	boolean east = false;
	boolean west = false;

	//Constructor
	public Hexagon(){
		this.ID = ID;
		this.centroidLatitude = centroidLatitude;
		this.centroidLongitude = centroidLongitude;
		this.k = k;
		this.M = M;
		this.F = F;
		this.sanctuary = sanctuary;
		this.sea = sea;
		//this.inIndividualsData = new HashMap<Hexagon, float[]>();
		//this.outIndividualsData = new HashMap<Hexagon, float[]>();
		this.biomass = biomass;
		this.north = north;
		this.south = south;
		this.east = east;
		this.west = west;
		this.isStation = false;
		this.sanctuary = false;
		//initilizate the later array that is going to be filled with the classes present on the hex. False by default
		initizlizateClassesPresence();

	}

	//first phase of the Phi function
	/**
	 * In this phase the agent has to calculate its neighborhood. Currently the hexagon performs 
	 * a spatial query, whether it checks the other hexagons around a distance from the hexagon centroid 
	 **/
	@ScheduledMethod(start = 1, interval = 0, priority = 1)
	public void BoundaryCondition(){
		Context context = ContextUtils.getContext(this);
		Geography<Hexagon> geography = (Geography)context.getProjection("Geography");
		//hexagon is the geometry
		Geometry hexagon = geography.getGeometry(this);
		GeographyWithin<Hexagon> within = new GeographyWithin<Hexagon>(geography, distance , this);
		if( within.query() != null ){
			FilterIterator filter = new FilterIterator(within.query().iterator(), new InstanceofPredicate(Hexagon.class));
			for (Iterator iter = filter; iter.hasNext();)  {
				Hexagon obj = (Hexagon)iter.next();
				neighbors.add(obj);
				idList.add(obj.ID);
				double[] centroidsNeigh = new double[2];
				centroidsNeigh[0] = obj.centroidLatitude;
				centroidsNeigh[1] = obj.centroidLongitude;
				//map with ID--CentroidsCoordinates[]
				centroids.put(obj.ID, centroidsNeigh);
			}

		}else{
			System.out.println("ERROR: Hexagon "+this.ID+" has not neighbors! ");
		}
		//Assign the direction to the neighbors 
		for(Hexagon neigh : neighbors){
			double calcLatitude = round(calculateDeltaCoor(this.centroidLatitude, neigh.centroidLatitude), precision);
			double calcLongitude = round(calculateDeltaCoor(this.centroidLongitude, neigh.centroidLongitude), precision);

			if(calcLatitude == nordLatitude && calcLongitude == nordLongitude){
				neighborDirections.put(neigh.ID, "North");
				neighboursDirections.put("N",neigh.ID);
			} else if(calcLatitude == nordEstLatitude && calcLongitude == nordEstLongitude){
				neighborDirections.put(neigh.ID, "NorthEast");
				neighboursDirections.put("NE",neigh.ID);
			} else if(calcLatitude == nordWestLatitude && calcLongitude == nordWestLongitude){
				neighborDirections.put(neigh.ID, "NorthWest");
				neighboursDirections.put("NW",neigh.ID);
			} else if(calcLatitude == southLatitude && calcLongitude == southLongitude){
				neighborDirections.put(neigh.ID, "South");
				neighboursDirections.put("S",neigh.ID);
			} else if(calcLatitude == southEstLatitude && calcLongitude == southEstLongitude){
				neighborDirections.put(neigh.ID, "SouthEast");
				neighboursDirections.put("SE",neigh.ID);
			} else if(calcLatitude == southWestLatitude && calcLongitude == southWestLongitude){
				neighborDirections.put(neigh.ID, "SouthWest");
				neighboursDirections.put("SW",neigh.ID);
			}else {
				System.out.println("ERROR: You should check your shapefile, it is having some features problems. Due"
						+ "to the hexagon cannot find its neigbours. " + neigh.getID() );

			}
		}

	}
	//The method just for once will look at the cohorts and the sole presence on the sea
	//So the age classes will be loaded following the shape file of presences. 
	@ScheduledMethod(start = 1 , interval = 0 , priority = ScheduleParameters.FIRST_PRIORITY)
	public void SettingTheSolePresence(){
		Context context = ContextUtils.getContext(this);
		IndexedIterable<Hexagon> iterator = context.getObjects(Hexagon.class);
		if (iterator != null) {
			for (Hexagon hexagon : iterator) {
				for (int i = 0; i < hexagon.classesPresent.length ; i++) {
					for (Integer key : hexagon.cohort.keySet()) {
						double[] population = hexagon.cohort.get(key);
						if (hexagon.classesPresent[i] == false) {
							population[i] = 0.0;
							setPopulationEmigratedInCohort(key, population[i], i);
						}
					}
				}
			}
		}

	}

	//This increments the Cohort age, so is the first thing that is executed at tick 
	@ScheduledMethod(start = 2 , interval = 2 , priority = ScheduleParameters.FIRST_PRIORITY)
	public void IncrementingCohortTime(){
		if (this.cohort != null) {
			for (Integer key : this.cohort.keySet()) {
				Integer cohortTime = this.cohortAge.get(key);
				int nextAge = cohortTime + 1;
				int year = key;
				//this.cohortAge.remove(key);
				this.cohortAge.put(year, nextAge);
				//this.cohortTime.put(key, );

			}
		}//else {
		//do nothing, wait that flux will carry new waves 
		//}
	}

	//Second phase of the Phi function
	/*
	 * In this phase is where the population growth has to be performed. 
	 * As the simulation starts at Nov 1st, the multiple mortality has to be implemented
	 * as part of it i will do in a different method in that way i can test it and just it. 
	 *  
	 * 
	 * */
	@ScheduledMethod(start = 2 , interval = 2 , priority = 5)
	public void Collision(){
		Context context = ContextUtils.getContext(this);
		IndexedIterable<Kronos> iter = context.getObjects(Kronos.class);
		Kronos kronos = iter.get(0);
		int month = kronos.getMonthFromTick();
		//Iterating throw the cohorts
		for (Integer key : this.cohortAge.keySet()) {
			double[] population = this.cohort.get(key);
			for (int i = 0; i < population.length - 1; i++) {
				if (population[i] > 0) {
					switch (i) {
					case 0: //change from age zero to one
						double[] ageZeroLength = Utils.age0Interval;
						double upperLimit = ageZeroLength[1];
						int age = this.cohortAge.get(key);
						double length = Utils.vonBenrtanlaffyF(age,	this.k);
						if (length > upperLimit) {
							//all the population changes at once
							double popToChange = population[0] * (growthVector[month]/100.0);
							population[1] += popToChange;
							if (population[0] - popToChange > 0) {
								population[0] = population[0] - popToChange;
							} else {
								population[0] = 0;

							}
						}
						//reinizialization of the variables 
						age = 0;
						length = 0;
						break;
					case 1: //change from age class one to two
						double[] ageOneLength = Utils.age1Interval;
						double upperOneLimit = ageOneLength[1];
						int ageOne = this.cohortAge.get(key);
						double lengthCohortOne = Utils.vonBenrtanlaffyF(ageOne, this.k);
						if (lengthCohortOne > upperOneLimit) {
							double popToChange = population[1] * (growthVector[month]/100.0);
							population[2] += popToChange;
							if (population[1] - popToChange > 0) {
								population[1] = population[1] - popToChange;
							} else {
								population[1] = 0;

							}
						}
						//reinizialization of the variables 
						ageOne = 0;
						lengthCohortOne = 0;
						break;

					case 2: //change from age class two to three
						double[] ageTwoLength = Utils.age2Interval;
						double upperTwoLimit = ageTwoLength[1];
						int ageTwo = this.cohortAge.get(key);
						double lengthCohortTwo = Utils.vonBenrtanlaffyF(ageTwo, this.k);
						if (lengthCohortTwo > upperTwoLimit) {
							double popToChange = population[2] * (growthVector[month]/100.0);							
							population[3] += popToChange;
							if (population[2] - popToChange > 0) {
								population[2] = population[2] - popToChange;

							} else {
								population[2] = 0;

							}
						}
						//reinizialization of the variables 
						ageTwo = 0;
						lengthCohortTwo = 0;
						break;

					case 3: //change from age class three to four
						double[] ageThreeLength = Utils.age3Interval;
						double upperThreeLimit = ageThreeLength[1];
						int ageThree = this.cohortAge.get(key);
						double lengthCohortThree = Utils.vonBenrtanlaffyF(ageThree, this.k);
						if (lengthCohortThree > upperThreeLimit) {
							double popToChange  = population[3] * (growthVector[month]/100.0);
							population[4] += popToChange;
							if (population[3] - popToChange > 0) {
								population[3] = population[3] - popToChange;
							} else {
								population[3] = 0;
							}
						}
						//reinizialization of the variables 
						ageThree = 0;
						lengthCohortThree = 0;
						break;
					case 4: //change from age class four to five plus
						double[] ageFourLength = Utils.age4Interval;
						double upperFourLimit = ageFourLength[1];
						int ageFour = this.cohortAge.get(key);
						double lengthCohortFour = Utils.vonBenrtanlaffyF(ageFour, this.k);
						if (lengthCohortFour > upperFourLimit) {
							double popToChange = population[4] * (growthVector[month]/100.0);
							population[5] += popToChange;
							if (population[4] - popToChange > 0) {
								population[4] = population[4] - popToChange;
							} else {
								population[4] = 0;
							}
						}
						//reinizialization of the variables 
						ageFour = 0;
						lengthCohortFour = 0;
						break;
					}
				}
			}
			if (population != this.cohort.get(key)) {
				this.cohort.put(key, population);
			}//else do not change the population in the group
		}

	}

	@ScheduledMethod(start = 26 , interval = 24 , priority = ScheduleParameters.FIRST_PRIORITY)
	public void cleaningCohorts(){
		Map<Integer , double[]> cohort = this.cohort;
		List<Integer> toBeDeleted = new ArrayList<>(); 
		for (Integer year : cohort.keySet()) {
			double[] population  = cohort.get(year);
			double countPopulation = 0.0;
			for (int i = 0; i < population.length; i++) {
				countPopulation += population[i];
			}
			//if the population is bigger than zero and less than one remove it
			if (countPopulation > 0 && countPopulation <1 ) {
				//this.cohort.remove(year);
				toBeDeleted.add(year);
			}
		}
		//removing from cohort
		for (Integer year: toBeDeleted) {
			this.cohort.remove(year);
		}

	}


	/*
	 * 	This method should be used to calculate and update the fishery mortality 
	 * 	for both parameters Rapid and Gill net 	
	 * 
	 * the division makes it first, in the formulae 
	 * apply it for the cohorts
	 * */
	@ScheduledMethod(start = 2 , interval = 2 , priority = 3)
	public void naturalMortality(){
		Context context = ContextUtils.getContext(this);
		IndexedIterable<Ares> iterator = context.getObjects(Ares.class);
		Ares ares = iterator.get(0);
		IndexedIterable<Kronos> tempoIter = context
				.getObjects(Kronos.class);
		Kronos kronos = tempoIter.get(0);
		int year = kronos.getYearFomTick();
		int month = kronos.getMonthFromTick();
		Map<Integer, double[]> cohort = this.cohort;
		StringBuffer yearMonthKeySB = new StringBuffer();
		if (month < 10) {
			yearMonthKeySB.append(year).append("0").append(month);
		} else {
			yearMonthKeySB.append(year).append(month);
		}
		int yearMonthKey = Integer.valueOf(yearMonthKeySB.toString());
		Map<Integer, float[]> naturalProbTable = ares.getNaturalMortality();
		if (naturalProbTable != null
				&& naturalProbTable.containsKey(yearMonthKey)) {
			//Gets the natural mortalities probabilites from the month year key
			float[] mProbabilities = naturalProbTable.get(yearMonthKey);
			//map to iterate in this.cohort
			//Map<Integer, double[]> cohort = this.cohort;
			for (Integer key : cohort.keySet()) {
				double[] population = cohort.get(key);
				for (int x = 0; x < population.length; x++) {
					if (population[x] > 0) {
						BinomialDistribution binomialDistribution = new BinomialDistribution((int) population[x], 1 - mProbabilities[x]);	
						double nextPopulationClass = binomialDistribution.sample();
						//killing by natural causes
						double dead = population[x] - nextPopulationClass;
						population[x] = nextPopulationClass;
					}
				}
				//cohort.remove(key);
				cohort.put(key, population);
			}

		} else {
			//obtaining data from last year available 
			StringBuffer yearLastYearKeySB = new StringBuffer();
			int lastYear = kronos.getLastYear();
			if (month < 10) {
				yearLastYearKeySB.append(lastYear).append("0")
				.append(month);
			} else {
				yearLastYearKeySB.append(lastYear).append(month);
			}
			yearMonthKey = Integer.valueOf(yearLastYearKeySB.toString());
			Map<Integer, float[]> lastYearNatProbTable = ares
					.getNaturalMortality();
			if (lastYearNatProbTable.containsKey(yearMonthKey)) {
				float[] mProbabilities = naturalProbTable.get(yearMonthKey);
				for (Integer key : cohort.keySet()) {
					double[] population = cohort.get(key);
					for (int x = 0; x < population.length; x++) {
						if (population[x] > 0) {
							BinomialDistribution distro = new BinomialDistribution((int) population[x], 1 - mProbabilities[x]);	
							double nextPopulation = distro.sample();
							//killing by natural causes
							double dead = population[x] - nextPopulation;
							population[x] = nextPopulation;
						}
					}
					//cohort.remove(key);
					cohort.put(key, population);
				}

			} else {
				System.out
				.println("ERROR: when you dont have data even the key with the last year is wrong!");
			}
		}

		//copying cohort changes
		try {
			for (Integer yearKey : cohort.keySet()) {
				double[] population = cohort.get(yearKey);
				this.cohort.put(yearKey, population);
			}
		} catch (ConcurrentException exp) {
			// TODO: handle exception
			System.out.println("ERROR: Concurrent exception on:" + this.ID);
		}

	}

	@ScheduledMethod(start = 2 , interval = 2 , priority = 4)
	public void fishingMortatility(){
		if (this.isSanctuary() != true) {
			Context context = ContextUtils.getContext(this);
			IndexedIterable<Ares> iterator = context.getObjects(Ares.class);
			Ares ares = iterator.get(0);
			IndexedIterable<Kronos> tempoIter = context
					.getObjects(Kronos.class);
			Kronos kronos = tempoIter.get(0);
			int year = kronos.getYearFomTick();
			int month = kronos.getMonthFromTick();
			StringBuffer yearMonthKeySB = new StringBuffer();
			if (month < 10) {
				yearMonthKeySB.append(year).append("0").append(month);
			} else {
				yearMonthKeySB.append(year).append(month);
			}
			int yearMonthKey = Integer.valueOf(yearMonthKeySB.toString());
			double deathCountRapid = 0;
			double deathCountGillItaly = 0;
			double deathCountGillCroatia = 0;
			Map<Integer, double[]> cohort = this.cohort;
			if (this.getSea() == "IT") {

				//applying italian rapid fishery 
				Map<Integer, float[]> rapidFisheryTable = ares.getRapidItaly();
				if (rapidFisheryTable != null
						&& rapidFisheryTable.containsKey(yearMonthKey)) {
					float[] rapidFProbabilities = rapidFisheryTable
							.get(yearMonthKey);
					for (Integer key : cohort.keySet()) {
						double[] population = cohort.get(key);
						for (int x = 0; x < population.length; x++) {
							if (population[x] > 0) {
								//killing by rapid fishery
								BinomialDistribution distribution = new BinomialDistribution(
										(int) population[x],
										1 - (rapidFProbabilities[x] ));//this is for double fishery 
								double nextPopulation = distribution.sample();
								double dead = population[x] - nextPopulation;
								population[x] = nextPopulation;
								deathCountRapid += dead;
							}
						}
						cohort.put(key, population);
						this.rapidItalyCatches = deathCountRapid;
					}
				}

				//applying gill net fishery 
				Map<Integer, float[]> gillNetFisheryTable = ares
						.getGillNetItaly();
				if (gillNetFisheryTable != null
						&& gillNetFisheryTable.containsKey(yearMonthKey)) {
					float[] gillNetProbabilities = gillNetFisheryTable
							.get(yearMonthKey);
					for (Integer key : cohort.keySet()) {
						double[] population = cohort.get(key);
						for (int x = 0; x < population.length; x++) {
							if (population[x] > 0) {
								BinomialDistribution distro = new BinomialDistribution(
										(int) population[x],
										1 - (gillNetProbabilities[x] ));
								double nextPopulation = distro.sample();
								//killing by gill net fishery
								double dead = population[x] - nextPopulation;
								population[x] = nextPopulation;
								deathCountGillItaly += dead;
							}
						}
						cohort.put(key, population);
						this.gillItalyCatches = deathCountGillItaly;
					}

				}	

			} else {
				//applying croatian fishery 
				Map<Integer, float[]> gillNetFisheryTable = ares
						.getGillNetCroatia();
				if (gillNetFisheryTable != null
						&& gillNetFisheryTable.containsKey(yearMonthKey)) {
					float[] gillNetProbabilities = gillNetFisheryTable
							.get(yearMonthKey);
					for (Integer key : cohort.keySet()) {
						double[] population = cohort.get(key);
						for (int x = 0; x < population.length; x++) {
							if (population[x] > 0) {
								BinomialDistribution distro = new BinomialDistribution(
										(int) population[x],
										1 - (gillNetProbabilities[x] ));
								double nextPopulation = distro.sample();
								//killing by gill net fishery
								double dead = population[x] - nextPopulation;
								population[x] = nextPopulation;
								deathCountGillCroatia += dead;
							}

						}

						cohort.put(key, population);
						this.gillCroatiaCatches = deathCountGillCroatia;
					}

				}
			}
			
			//Applying cohort changes
			try {
				for (Integer yearKey : cohort.keySet()) {
					double[] population = cohort.get(yearKey);
					this.cohort.put(yearKey, population);
				}
			} catch (ConcurrentException exp) {
				// TODO: handle exception
				System.out.println("ERROR: Concurrent exception on:" + this.ID);
			}
		}
	}


	/*
	 * Third phase of the phi function. 
	 * In this phase we have to perform the propagation of the fish population. 
	 * To do it, the idea is use the migration vectors that depends of the Age class and position and time. 
	 * */
	@ScheduledMethod(start = 3, interval = 2 , priority = ScheduleParameters.LAST_PRIORITY)
	public void Propagation(){
		TupleSpace messenger = new TupleSpace();
		Context context = ContextUtils.getContext(this);
		IndexedIterable<Emigrator> iterator = context.getObjects(Emigrator.class);
		Emigrator emigrator = (Emigrator) iterator.get(0);

		Map<String, double[]> migrationVectors = emigrator.getRules();
		Map<Integer, double[]> cohort = this.cohort;

		String location = null;
		if(this.north && this.east){
			location = "NE";
		} else if (this.north && this.west) {
			location = "NW";
		} else if (this.south && this.east) {
			location = "SE";
		} else if (this.south && this.west) {
			location = "SW";
		} else {
			System.out.println("ERROR: this occures when the location (variables North, South, West, East) from the shapefile was not loaded");
		}

		for (Integer key : cohort.keySet()) {
			double[] population = cohort.get(key);
			for (int i = 0; i < population.length; i++) {
				if (population[i] > 0.0) {
					String code = location + "-" + i;
					double[] vector = migrationVectors.get(code);
					if (vector != null) {
						double moveProbability = 0.0;
						for (int j = 0; j < vector.length; j++) {
							moveProbability += vector[j];
						}
						if (moveProbability <= 100) {
							BinomialDistribution distro = new BinomialDistribution(
									(int) population[i],(moveProbability / 100));
							double populationSample = distro.sample();
							double popThatMigratesDirections = 0.0;

							if (populationSample > 0.0) {

								//just controlling the population remains positive
								if (population[i] - populationSample > 0.0) {
									//update the population
									double newPopulation = population[i] - populationSample;
									setPopulationEmigratedInCohort(key, newPopulation, i);
								}
								//distributing the population sample on the directions
								popThatMigratesDirections = populationSample;
								for (int j = 0; j < vector.length; j++) {
									if (vector[j] != 0.0) {
										//here we have to distribute the population on the neighbors 
										switch (j) {
										case 0:
											if (neighboursDirections.containsKey("N")) {
												Integer idNeighborNorth = neighboursDirections.get("N");
												double moveProbabilityProportion0 = (vector[j] / moveProbability) *100;
												double populationProportion0 = (popThatMigratesDirections * moveProbabilityProportion0) /100;
												for (Hexagon neighbor : neighbors) {
													if (idNeighborNorth == (Integer) neighbor.ID) {
														if (neighbor.getPopulationPresence(i) == true) {
															messenger.postValue(neighbor, key, populationProportion0, i);

														}else {
															setPopulationEmigratedInCohort(key, populationProportion0, i);
														}
													}

												}
											}

											break;
										case 1:
											if (neighboursDirections.containsKey("NE")) {
												Integer idNeighborNorthEast = neighboursDirections.get("NE");
												double moveProportion1 = (vector[j] / moveProbability ) * 100;
												double populationProportion1 = (popThatMigratesDirections * moveProportion1) /100;
												for (Hexagon neighbor : neighbors) {
													if (idNeighborNorthEast == neighbor.ID) {
														if (neighbor.getPopulationPresence(i) == true) {
															messenger.postValue(neighbor, key, populationProportion1, i);
														}else {
															setPopulationEmigratedInCohort(key, populationProportion1, i);
														}
													}
												}
											}
											break;
										case 2:
											if (neighboursDirections.containsKey("NW")) {
												Integer idNeighborNorthWest = neighboursDirections.get("NW");
												double moveProportion2 = (vector[j]	/ moveProbability) *100;
												double populationProportion2 = (popThatMigratesDirections * moveProportion2)/100;
												for (Hexagon neighbor : neighbors) {
													if (idNeighborNorthWest == (Integer) neighbor.ID) {
														if (neighbor.getPopulationPresence(i) == true) {
															messenger.postValue(neighbor, key, populationProportion2, i);
														}else {
															setPopulationEmigratedInCohort(key, populationProportion2, i);
														}
													}
												}
											}
											break;
										case 3:
											if (neighboursDirections.containsKey("S")) {
												Integer idNeighborSouth = neighboursDirections.get("S");
												double moveProportion3 = (vector[j] / moveProbability) * 100;
												double populationProportion3 = (popThatMigratesDirections * moveProportion3)/100;
												for (Hexagon neighbor : neighbors) {
													if (idNeighborSouth == (Integer) neighbor.ID) {
														if (neighbor.getPopulationPresence(i) == true) {
															messenger.postValue(neighbor, key, populationProportion3,i);
														}else {
															setPopulationEmigratedInCohort(key, populationProportion3, i);
														}
													}

												}
											}
											break;
										case 4:
											if (neighboursDirections.containsKey("SE")) {
												Integer idNeighborSouthEast = neighboursDirections.get("SE");
												double moveProportion4 = (vector[j] / moveProbability) * 100;
												double populationProportion4 = (popThatMigratesDirections * moveProportion4)/100;
												for (Hexagon neighbor : neighbors) {
													if (idNeighborSouthEast == (Integer) neighbor.ID) {
														if (neighbor.getPopulationPresence(i) == true) {
															messenger.postValue(neighbor, key, populationProportion4, i);
														}else {
															setPopulationEmigratedInCohort(key, populationProportion4, i);
														}
													}

												}
											}
											break;
										case 5:
											if (neighboursDirections.containsKey("SW")) {
												Integer idNeighborSouthWest = neighboursDirections.get("SW");
												double moveProportion5 = (vector[j] / moveProbability) * 100;
												double populationProportion5 = (popThatMigratesDirections * moveProportion5)/100;
												for (Hexagon neighbor : neighbors) {
													if (idNeighborSouthWest == (Integer) neighbor.ID) {
														if (neighbor.getPopulationPresence(i) == true) {
															messenger.postValue(neighbor, key, populationProportion5, i);
														}else {
															setPopulationEmigratedInCohort(key, populationProportion5, i);
														}
													}

												}
											}
											break;

										}

									}

								}
							}
						} else {
							System.out
							.println("ERROR: The probability vector exceds the 100%. key: "	+ code);
						}
					} else {
						System.out
						.println("ERROR: Migration vector returns null. Probably bad code formation.");
					}
				}				
			}
		}
	}

	@ScheduledMethod(start = 2, interval = 1 , priority = ScheduleParameters.LAST_PRIORITY)
	public void biomassCalculus(){
		for (Integer key : this.cohort.keySet()) {
			double[] population = this.cohort.get(key);
			for (int i = 0; i < population.length && population[i] > 0; i++) {	
				switch (i) {
				case 0:
					this.biomass[0] = 0.0;
					double[] lengthInterval = Utils.age0Interval;
					for (int j = 0; j < population[0]; j++) {
						double length = RandomHelper.nextDoubleFromTo(
								lengthInterval[0], lengthInterval[1]);
						double weight = Utils.getWeightFromLength(length);
						this.biomass[0] += weight;
					}
					break;
				case 1:
					this.biomass[1] = 0.0;
					double[] lengthIntervalOne = Utils.age1Interval;
					for (int j = 0; j < population[1]; j++) {
						double length = RandomHelper.nextDoubleFromTo(
								lengthIntervalOne[0], lengthIntervalOne[1]);
						double weight = Utils.getWeightFromLength(length);
						this.biomass[1] += weight;
					}
					break;
				case 2:
					double[] lengthIntervalTwo = Utils.age2Interval;
					for (int j = 0; j < population[0]; j++) {
						double length = RandomHelper.nextDoubleFromTo(
								lengthIntervalTwo[0], lengthIntervalTwo[1]);
						double weight = Utils.getWeightFromLength(length);
						this.biomass[2] += weight;
					}
					break;
				case 3:
					this.biomass[3] = 0.0;
					double[] lengthIntervalThree = Utils.age3Interval;
					for (int j = 0; j < population[0]; j++) {
						double length = RandomHelper.nextDoubleFromTo(
								lengthIntervalThree[0], lengthIntervalThree[1]);
						double weight = Utils.getWeightFromLength(length);
						this.biomass[3] += weight;
					}
					break;
				case 4:
					this.biomass[4] = 0.0;
					double[] lengthIntervalFour = Utils.age4Interval;
					for (int j = 0; j < population[0]; j++) {
						double length = RandomHelper.nextDoubleFromTo(
								lengthIntervalFour[0], lengthIntervalFour[1]);
						double weight = Utils.getWeightFromLength(length);
						this.biomass[4] += weight;
					}
					break;
				case 5:
					this.biomass[5] = 0.0;
					double[] lengthIntervalFive = Utils.age5Interval;
					for (int j = 0; j < population[0]; j++) {
						double length = RandomHelper.nextDoubleFromTo(
								lengthIntervalFive[0], lengthIntervalFive[1]);
						double weight = Utils.getWeightFromLength(length);
						this.biomass[5] += weight;
					}
					break;
				}
			}
		}
	}

	private void initizlizateClassesPresence() {
		for(int x = 0; x < this.classesPresent.length; x++){
			this.classesPresent[x]= false;
		}

	}

	public double getCentroidLatitude() {
		return centroidLatitude;
	}
	public void setCentroidLatitude(double centroidLatitude) {
		this.centroidLatitude = centroidLatitude;
	}
	public double getCentroidLongitude() {
		return centroidLongitude;
	}
	public void setCentroidLongitude(double centroidLongitude) {
		this.centroidLongitude = centroidLongitude;
	}
	public double getK() {
		return k;
	}
	public void setK(double k) {
		this.k = k;
	}
	public double getM() {
		return M;
	}
	public void setM(double m) {
		M = m;
	}
	public double getF() {
		return F;
	}

	public void setF(double fProbability){
		this.F = fProbability;
	}
	public int getID() {
		return this.ID;
	}
	public void setID(int iD) {
		this.ID = iD;
	}

	public double[] getPopulation() {
		return population;
	}

	public void setPopulation(double[] population) {

		this.population = population;
	}

	public void setPopulation(double population, int classAge){
		this.population[classAge] += population;
	}

	public boolean[] getClassesPresent() {
		return classesPresent;
	}

	public boolean getPopulationPresence(int type){
		return this.classesPresent[type];
	}

	public void setClassesPresent(boolean[] classesPresent) {
		this.classesPresent = classesPresent;
	}

	public void setClassesPresent(int presence){
		this.classesPresent[presence]= true;
	}

	public boolean isSanctuary() {
		return sanctuary;
	}

	public void setSanctuary(boolean sanctuary) {
		this.sanctuary = sanctuary;
	}

	public boolean isRapidPresence() {
		return rapidPresence;
	}

	public void setRapidPresence(boolean rapidPresence) {
		this.rapidPresence = rapidPresence;
	}

	public boolean isGillPresence() {
		return gillPresence;
	}

	public void setGillPresence(boolean gillPresence) {
		this.gillPresence = gillPresence;
	}

	public String getSea() {
		return this.sea;
	}

	public void setSea(String sea) {
		this.sea = sea;
	}

	public double calculateDeltaCoor(double coordOne, double coordTwo){
		return coordOne - coordTwo;
	}

	public static double round(final double number, final int precision){
		return Math.round(number * Math.pow(10, precision)) / Math.pow(10, precision);
	}

	//change it and consider now that are cohorts.... 
	public double getClass0Population(){
		double popClassZero = 0.0;
		for (Integer key : this.cohort.keySet()) {
			popClassZero += this.cohort.get(key)[0];

		}
		return popClassZero;
	}

	public double getClass1Population(){
		double popClassOne = 0.0;
		for (Integer key : this.cohort.keySet()) {
			popClassOne += this.cohort.get(key)[1];

		}
		return popClassOne;
	}

	public double getClass2Population(){
		double popClassTwo = 0.0;
		for (Integer key : this.cohort.keySet()) {
			popClassTwo += this.cohort.get(key)[2];

		}
		return popClassTwo;
	}

	public double getClass3Population(){
		double popClassThree = 0.0;
		for (Integer key : this.cohort.keySet()) {
			popClassThree += this.cohort.get(key)[3];

		}
		return popClassThree;
	}

	public double getClass4Population(){
		double popClassFour = 0.0;
		for (Integer key : this.cohort.keySet()) {
			popClassFour += this.cohort.get(key)[4];

		}
		return popClassFour;
	}
	public double getClass5Population(){
		double popClassFive = 0.0;
		for (Integer key : this.cohort.keySet()) {
			popClassFive += this.cohort.get(key)[5];

		}
		return popClassFive;
	}

	public double getNumberOfIndividuals(){
		double popTotal = 0.0;
		for (Integer key : this.cohort.keySet()) {
			double[] population = this.cohort.get(key);
			for (int i = 0; i < population.length; i++) {
				popTotal += population[i];
			}

		}
		return popTotal;
	}

	public double getBiomass() {
		double biomassValue = 0;
		for (double value : biomass){
			biomassValue+= value;
		}
		return biomassValue;
	}

	public void setBiomass(double[] biomass) {
		this.biomass = biomass;
	}

	public double getRapidItalyCatches() {
		return rapidItalyCatches;
	}

	public void setRapidItalyCatches(double rapidItalyCatches) {
		this.rapidItalyCatches = rapidItalyCatches;
	}

	public double getGillItalyCatches() {
		return gillItalyCatches;
	}

	public void setGillItalyCatches(double gillItalyCatches) {
		this.gillItalyCatches = gillItalyCatches;
	}

	public double getGillCroatiaCatches() {
		return gillCroatiaCatches;
	}

	public void setGillCroatiaCatches(double gillCroatiaCatches) {
		this.gillCroatiaCatches = gillCroatiaCatches;
	}

	public Map<Integer, double[]> getCohort() {
		return cohort;
	}

	public void setCohort(Map<Integer, double[]> cohort) {
		this.cohort = cohort;
	}

	public void setPopulationCohort(Integer year, double[] population){
		this.cohort.put(year, population);
	}

	public void setPopulationEmigratedInCohort(Integer year, double emigratedPopulation, int age){
		if (this.cohort.containsValue(year)) {
			double[] population = this.cohort.get(year);
			population[age] += emigratedPopulation;
			this.cohort.put(year, population);
		}
	}

	public Map<Integer, Integer> getCohortAge() {
		return cohortAge;
	}

	public void setNewCohortTime(Integer key){
		//the age is because is the number of months 
		//that has passed since the month in which where born 
		//and the time that the survey was performed-- review this number
		this.cohortAge.put(key, 9);
	}

	public void setCohortTimeInitial(Integer key) {
		this.cohortAge.put(key, 0);
	}

	public boolean isNorth() {
		return north;
	}

	public void setNorth(boolean north) {
		this.north = north;
	}

	public boolean isSouth() {
		return south;
	}

	public void setSouth(boolean south) {
		this.south = south;
	}

	public boolean isEast() {
		return east;
	}

	public void setEast(boolean east) {
		this.east = east;
	}

	public boolean isWest() {
		return west;
	}

	public void setWest(boolean west) {
		this.west = west;
	}

	public boolean isStation() {
		return isStation;
	}

	public void setStation(boolean isStation) {
		this.isStation = isStation;
	}

	//for testing purpose
	//@ScheduledMethod(start = 3 , interval = 24, priority = ScheduleParameters.FIRST_PRIORITY)
	public void dataCollector(){
		System.out.println("tick; Station; latitude; longitude; ");
		if (this.isStation) {
			//change here
			double tick = RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
			double numberOfIndividuals = this.getNumberOfIndividuals();
			double class0Population = this.getClass0Population();
			double class1Population = this.getClass1Population();
			double class2Population = this.getClass2Population();
			double class3Population = this.getClass3Population();
			double class4Population = this.getClass4Population();
			double class5Population = this.getClass5Population();
			double biomass = this.getBiomass();
			double latStation = this.latStation;
			double longStation = this.longStation;
			System.out.println( tick +";" + this.getID() + latStation+ ";"  + longStation + ";" );
		}//else do nothing

	}

	@ScheduledMethod(start = 24, interval = 24, priority = ScheduleParameters.LAST_PRIORITY)
	public void cleanStation(){
		if (this.isStation){
			this.isStation = false;
		}
	}
	
	public double getLatStation() {
		return latStation;
	}

	public void setLatStation(double latStation) {
		this.latStation = latStation;
	}

	public double getLongStation() {
		return longStation;
	}

	public void setLongStation(double longStation) {
		this.longStation = longStation;
	}

	public int getNameStation() {
		return nameStation;
	}

	public void setNameStation(int nameStation) {
		this.nameStation = nameStation;
	}


}