package micro;

import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.jgoodies.common.base.SystemUtils;

import cern.jet.random.Uniform;
import repast.simphony.context.Context;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.parameter.Parameter;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.Grid;
import repast.simphony.util.ContextUtils;

/**
 * The agent for the Solea model.
 * 
 * @author
 */

public class Agent {

	private int currentAge, ageClass;
	private double length, weight;
	//parameter to change along the simulation time
	private boolean dead = false;
	public int hexagonID;

	public Agent(int ageClass) {
		this.ageClass = ageClass;
		float[] lengthRange = ParametersManager.getClassLengthData().get(
				this.ageClass);
		this.length = RandomHelper.nextDoubleFromTo(lengthRange[0],
				lengthRange[1]);
		this.currentAge = (int) this.getAgeFromLength(this.length);
		this.weight = this.getWeightFromLength(this.length);

	}

	public Agent(int ageClass, int age) {

		this.ageClass = ageClass;
		System.out.println("age: " + this.ageClass);
		this.currentAge = age;
		System.out.println("Current age: " + this.currentAge);
		this.length = this.getLengthFromAge(age);
		this.weight = this.getWeightFromLength(this.length);
	}

	// Schedule the step method for agents. The method is scheduled starting at
	// tick one with an interval of 1 tick. Specifically, the step starts at 0,
	// and
	// and recurs at 1,2,3,...etc
	//@ScheduledMethod(start = 1, interval = 1)
	public void step() {
		// increase my age
		currentAge++;
		this.length = this.getLengthFromAge(this.currentAge);
		this.weight = this.getWeightFromLength(this.length);
		this.ageClass = this.getClassFromLength(this.length);
	}


	// Probabilistically kill the agent for natural factors
	//@ScheduledMethod(start = 2, interval = 1, priority = ScheduleParameters.LAST_PRIORITY)
	public void dieByNaturalProbability(){
		float prob;
		Map<Integer, float[]> naturalMortalityTable = ParametersManager
				.getNaturalMortalityData().get(Utils.getYearFromTick());
		if (naturalMortalityTable == null)
			naturalMortalityTable = ParametersManager.getNaturalMortalityData()
			.get(ParametersManager.getFirstYear()
					+ ParametersManager.getNumberOfYears() - 1);
		prob = naturalMortalityTable.get(this.ageClass)[this.currentAge % 12];
		//System.out.println("prob natural: " + prob +" age:" +this.ageClass + " year:"+ Utils.getYearFromTick());
		//natural probability of dead 
		if (Math.random() <= prob){
			dead = true;
		}

		if (dead == true) {
			// Get the context in which the agent resides.
			Context<Object> context = (Context) ContextUtils.getContext(this);
			// Remove the agent from the context
			context.remove(this);
		}

	}

	// Probabilistically kill the agent by fishery 
	//@ScheduledMethod(start = 2, interval = 1, priority = ScheduleParameters.LAST_PRIORITY)
	public void dieByFishingProbability(){
		float prob;
		Context<Object> context = (Context) ContextUtils.getContext(this);
		if (context != null){
			if (!ParametersManager.isVirginStock() && !dead) {
				Map<Integer, float[]> fishingMortalityTable = ParametersManager
						.getFishingMortality().get(Utils.getYearFromTick());
				if (fishingMortalityTable == null)
					fishingMortalityTable = ParametersManager.getFishingMortality()
					.get(ParametersManager.getFirstYear()
							+ ParametersManager.getNumberOfYears() - 1);
				prob = fishingMortalityTable.get(this.ageClass)[this.currentAge % 12];
				if (Math.random() <= prob)
					dead = true;

			}
			if (dead) {
				//remove agent from context
				context.remove(this);
			}
		}

	}

	public int getCurrentAge() {
		return currentAge;
	}

	public int getAgeClass() {
		return ageClass;
	}

	public double getLength() {
		return length;
	}

	public double getWeight() {
		return weight;
	}

	public int isClass0() {
		return this.ageClass == 0 ? 1 : 0;
	}

	public int isClass1() {
		return this.ageClass == 1 ? 1 : 0;
	}

	public int isClass2() {
		return this.ageClass == 2 ? 1 : 0;
	}

	public int isClass3() {
		return this.ageClass == 3 ? 1 : 0;
	}

	public int isClass4() {
		return this.ageClass == 4 ? 1 : 0;
	}

	public int isClass5() {
		return this.ageClass >= 5 ? 1 : 0;
	}

	// fVB inverse
	private double getAgeFromLength(double l) {	
		Map<Integer, float[]> kvbfTable = ParametersManager
				.getKData().get(Utils.getYearFromTick());
		if (kvbfTable == null)
			kvbfTable = ParametersManager.getKData()
			.get(ParametersManager.getFirstYear()
					+ ParametersManager.getNumberOfYears() - 1);
		float kvbf = kvbfTable.get(this.ageClass)[this.currentAge % 12];

		double formula = 12 * (ParametersManager.getT_0() - (Math.log(1 - l
				/ ParametersManager.getL_inf())) / kvbf);
		return formula ;
	}

	// fVB
	private double getLengthFromAge(double a) {
		Map<Integer, float[]> kvbfTable = ParametersManager
				.getKData().get(Utils.getYearFromTick());
		if (kvbfTable == null)
			kvbfTable = ParametersManager.getKData()
			.get(ParametersManager.getFirstYear()
					+ ParametersManager.getNumberOfYears() - 1);


		float kvbf = kvbfTable.get(this.ageClass)[this.currentAge % 12];
		//System.out.println("agente age: " + this.currentAge %12+ " k: "+ kvbf+ " year: " + Utils.getYearFromTick());
		double nextLength = ParametersManager.getL_inf()
				* (1 - Math.exp(-kvbf
						* (a / 12 - ParametersManager.getT_0())));
		//System.out.println("current lenght: " +this.length + " next lenght: " + nextLength );
		return nextLength ;
	}

	private double getWeightFromLength(double l) {
		return ParametersManager.getA_length_weight()
				* Math.pow(l, ParametersManager.getB_length_weight());
	}

	private int getClassFromLength(double l) {
		Set<Entry<Integer, float[]>> lengthRange = ParametersManager
				.getClassLengthData().entrySet();
		float[] temp;
		Iterator<Entry<Integer, float[]>> it = lengthRange.iterator();
		Entry<Integer, float[]> currentEntry;
		while (it.hasNext()) {
			currentEntry = it.next();
			temp = currentEntry.getValue();
			if (l > temp[0] && l <= temp[1])
				return currentEntry.getKey();
		}
		return ParametersManager.getNumberOfAges() - 1;
	}

	public double getAppearanceLength() {
		return this.length / 1000;
	}

	public int isSolea() {
		return 1;
	}



}
