 package micro;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.HashMap;
import java.util.Map;

public class InputParser {

	private String path;
	private int firstYear;
	private int numOfAges;
	private int numOfYears;
	// (year,[class0,class1,...])
	private  Map<Integer, float[]> indexSurvey;

	private double class0stdev;
	// (class,[min,max])
	private Map<Integer, float[]> classLengthRange;
	// (class,[Jan,Feb,...])
	private Map<Integer, Map<Integer, float[]>> naturalMortality;

	// (year,(class,[Jan,Feb,...]))
	private Map<Integer, Map<Integer, float[]>> kvbf;

	public double getClass0stdev() {
		return class0stdev;
	}

	// (year,(class,[Jan,Feb,...]))
	private Map<Integer, Map<Integer, float[]>> fishingMortality;

	public InputParser(String path) {
		this.class0stdev=0;
		this.path = path;
		this.indexSurvey = new HashMap<Integer, float[]>();
		this.classLengthRange = new HashMap<Integer, float[]>();
		this.naturalMortality = new HashMap<Integer, Map<Integer, float[]>>();
		this.fishingMortality = new HashMap<Integer, Map<Integer, float[]>>();
		this.kvbf = new HashMap<Integer, Map<Integer, float[]>>();
		this.parse();
	}

	private void parse() {
		LineNumberReader r = null;
		try {
			r = new LineNumberReader(new FileReader(this.path));
			String line;
			while ((line = r.readLine()) != null) {
				if (line.startsWith("%First year")) {
					this.firstYear = Integer.parseInt(r.readLine());
				} else if (line.startsWith("%Number of ages")) {
					this.numOfAges = Integer.parseInt(r.readLine());
				} else if (line.startsWith("%Number of years")) {
					this.numOfYears = Integer.parseInt(r.readLine());
				} else if (line.startsWith("%Survey Index")) {
					this.parseIndexSurvey(r);
				} else if (line.startsWith("%Age Classes Length")) {
					this.parseClassLengthRange(r);
				} else if (line
						.startsWith("%Monthly probability of natural mortality")) {
					this.parseNaturalMortality(r);
				} else if (line
						.startsWith("%Monthly probability of fishing mortality")) {
					this.parseFishingMortality(r);
				}
				else if (line.startsWith("%k")) {
					this.parseKvbf(r);
				}
			}
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NumberFormatException e) {
			System.out.println("Generic parsing error");
			e.printStackTrace();
		}
		try {
			r.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public void setIndexSurvey(Map<Integer, float[]> indexSurvey) {
		this.indexSurvey = indexSurvey;
	}

	private void parseIndexSurvey(LineNumberReader r) {
		String line = null;
		String[] field;

		try {
			while ((line = r.readLine()) != null && !line.startsWith("%")) {
				if (!line.startsWith("#")
						&& line.replaceAll("\\s+", "").length() > 0) {
					field = line.split(";");
					float[] f = new float[this.numOfAges];
					for (int i = 0; i < f.length && i < field.length - 1; i++) {
						f[i] = Float.parseFloat(field[i + 1]);
					}
					this.indexSurvey.put(Integer.parseInt(field[0]), f);
				}
				r.mark(1024);
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// r.setLineNumber(r.getLineNumber() - 1);
		try {
			r.reset();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		float sum=0;
		for (int i = 0; i < this.indexSurvey.size(); i++) {
			sum+=this.indexSurvey.get(this.firstYear+i)[0];
		}
		float avg=sum/this.indexSurvey.size();
		float sqresiduals=0;
		for (int i = 0; i < this.indexSurvey.size(); i++) {
			sqresiduals+=Math.pow(this.indexSurvey.get(this.firstYear+i)[0]- avg, 2);
		}
		this.class0stdev= Math.sqrt(sqresiduals / this.indexSurvey.size());
	}

	private void parseFishingMortality(LineNumberReader r) {
		String line = null;
		String[] field;
		try {
			while ((line = r.readLine()) != null && !line.startsWith("%")) {
				if (!line.startsWith("#")
						&& line.replaceAll("\\s+", "").length() > 0) {
					field = line.split(";");
					float[] f = new float[12];
					for (int i = 0; i < f.length && i < field.length - 2; i++) {
						f[i] = Float.parseFloat(field[i + 2]);
					}
					Map<Integer, float[]> tmp = this.fishingMortality
							.get(Integer.parseInt(field[1]));
					if (tmp == null)
						tmp = new HashMap<Integer, float[]>();
						//be careful with the parsing type (int, float) 
						tmp.put(Integer.parseInt(field[0]), f);
						this.fishingMortality.put(Integer.parseInt(field[1]), tmp);
				}
				r.mark(1024);
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// r.setLineNumber(r.getLineNumber() - 1);
		try {
			r.reset();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private void parseClassLengthRange(LineNumberReader r) {
		String line = null;
		String[] field;
		try {
			while ((line = r.readLine()) != null && !line.startsWith("%")) {
				if (!line.startsWith("#")
						&& line.replaceAll("\\s+", "").length() > 0) {
					field = line.split(";");
					float[] f = new float[2];
					for (int i = 0; i < f.length && i < field.length - 1; i++) {
						f[i] = Float.parseFloat(field[i + 1]);
					}
					this.classLengthRange.put(Integer.parseInt(field[0]), f);
				}
				r.mark(1024);
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// r.setLineNumber(r.getLineNumber() - 1);
		try {
			r.reset();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private void parseNaturalMortality(LineNumberReader r) {
		String line = null;
		String[] field;
		try {
			while ((line = r.readLine()) != null && !line.startsWith("%")) {
				if (!line.startsWith("#")
						&& line.replaceAll("\\s+", "").length() > 0) {
					field = line.split(";");
					float[] f = new float[12];
					for (int i = 0; i < f.length && i < field.length - 2; i++) {
						f[i] = Float.parseFloat(field[i + 2]);
					}
					Map<Integer, float[]> tmp = this.naturalMortality
							.get(Integer.parseInt(field[1]));
					if (tmp == null)
						tmp = new HashMap<Integer, float[]>();
						tmp.put(Integer.parseInt(field[0]), f);
						this.naturalMortality.put(Integer.parseInt(field[1]), tmp);
				}
				r.mark(1024);
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// r.setLineNumber(r.getLineNumber() - 1);
		try {
			r.reset();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private void parseKvbf(LineNumberReader r) {
		String line = null;
		String[] field;
		try {
			while ((line = r.readLine()) != null && !line.startsWith("%")) {

				if (!line.startsWith("#")
						&& line.replaceAll("\\s+", "").length() > 0) {
					field = line.split(";");
					float[] f = new float[12];
					for (int i = 0; i < f.length && i < field.length - 2; i++) {
						f[i] = Float.parseFloat(field[i + 2]);
					}
					Map<Integer, float[]> tmp = this.kvbf
							.get(Integer.parseInt(field[1]));
					if (tmp == null)
						tmp = new HashMap<Integer, float[]>();
						tmp.put(Integer.parseInt(field[0]), f);
						this.kvbf.put(Integer.parseInt(field[1]), tmp);
				}
				r.mark(1024);
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// r.setLineNumber(r.getLineNumber() - 1);
		try {
			r.reset();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public String getPath() {
		return path;
	}

	public int getFirstYear() {
		return firstYear;
	}

	public int getNumOfAges() {
		return numOfAges;
	}

	public int getNumOfYears() {
		return numOfYears;
	}

	public Map<Integer, float[]> getIndexSurvey() {
		return indexSurvey;
	}

	public Map<Integer, float[]> getClassLengthRange() {
		return classLengthRange;
	}

	public Map<Integer, Map<Integer, float[]>> getNaturalMortality() {
		return naturalMortality;
	}

	/*public Map<Integer, float[]> getKvbf() {
		return kvbf;
	}
	 */
	public Map<Integer, Map<Integer, float[]>> getFishingMortality() {
		return fishingMortality;
	}

	//change
	public Map<Integer, Map<Integer, float[]>> getKData(){
		return kvbf;
	}


	/*public void setKvbf(Map<Integer, float[]> kvbf) {
		this.kvbf = kvbf;
	}
	 */
	public static void main(String[] args) {
		//change the file for the current of pierluiggi
		InputParser p = new InputParser("misc/input DISPAS.csv");
		System.out.println(p.getFirstYear());

	}
}
